# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import re

from drgn import Object, offsetof, reinterpret
from drgn.commands import CommandError, CommandNotFoundError
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.sched import task_thread_info
from tests.linux_kernel import possible_cpus, skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestStruct(CrashCommandTestCase):
    def test_type(self):
        cmd = self.check_crash_command("struct list_head")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertRegex(cmd.stdout, "(?m)^SIZE: [0-9]+$")
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "struct list_head"
        )
        self.assertIsInstance(cmd.drgn_option.globals["size"], int)

    def test_typedef(self):
        # atomic_t is a typedef of an anonymous struct. Crash allows this.
        cmd = self.check_crash_command("struct atomic_t")
        self.assertIn("} atomic_t", cmd.stdout)
        self.assertIn('prog.type("atomic_t")', cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["type"].type_name(), "atomic_t")

    def test_not_found(self):
        self.assertRaises(
            LookupError, self.run_crash_command, "struct drgn_test_non_existent"
        )

    def test_not_found_drgn_option(self):
        cmd = self.run_crash_command_drgn_option(
            "struct drgn_test_non_existent", mode="compile"
        )
        self.assertIn('prog.type("struct drgn_test_non_existent")', cmd.stdout)

    @skip_unless_have_test_kmod
    def test_union(self):
        self.assertRaises(LookupError, self.run_crash_command, "struct drgn_test_union")

    @skip_unless_have_test_kmod
    def test_anonymous_union(self):
        self.assertRaises(
            LookupError, self.run_crash_command, "struct drgn_test_anonymous_union"
        )

    def test_type_member(self):
        cmd = self.check_crash_command("struct list_head.next")
        self.assertRegex(cmd.stdout, r"\[[0-9]+\] struct list_head \*next;")
        self.assertEqual(
            cmd.drgn_option.globals["next_type"].type_name(), "struct list_head *"
        )
        self.assertIsInstance(cmd.drgn_option.globals["next_offset"], int)

    def test_type_multiple_members(self):
        cmd = self.check_crash_command("struct list_head.next,prev")
        self.assertRegex(cmd.stdout, r"\[[0-9]+\] struct list_head \*next;")
        self.assertRegex(cmd.stdout, r"\[[0-9]+\] struct list_head \*prev;")
        self.assertEqual(
            cmd.drgn_option.globals["next_type"].type_name(), "struct list_head *"
        )
        self.assertIsInstance(cmd.drgn_option.globals["next_offset"], int)
        self.assertEqual(
            cmd.drgn_option.globals["prev_type"].type_name(), "struct list_head *"
        )
        self.assertIsInstance(cmd.drgn_option.globals["prev_offset"], int)

    @skip_unless_have_test_kmod
    def test_address(self):
        address = self.prog["drgn_test_singular_list"].address_
        cmd = self.check_crash_command(f"struct list_head {address:x}")
        self.assertIn("(struct list_head){", cmd.stdout)
        self.assertIn(
            'Object(prog, "struct list_head", address=0x', cmd.drgn_option.stdout
        )
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_singular_list"]
        )

    @skip_unless_have_test_kmod
    def test_symbol(self):
        cmd = self.check_crash_command("struct list_head drgn_test_singular_list")
        self.assertIn("(struct list_head){", cmd.stdout)
        self.assertIn(
            'object = prog["drgn_test_singular_list"]', cmd.drgn_option.stdout
        )
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_singular_list"]
        )

    @skip_unless_have_test_kmod
    def test_symbol_wrong_type(self):
        cmd = self.check_crash_command("struct hlist_head drgn_test_singular_list")
        self.assertIn("(struct hlist_head){", cmd.stdout)
        self.assertIn("prog.symbol", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"],
            reinterpret("struct hlist_head", self.prog["drgn_test_singular_list"]),
        )

    def test_member(self):
        cmd = self.check_crash_command("struct task_struct.pid init_task")
        self.assertRegex(cmd.stdout, r"pid = \([^)]*\)0")
        self.assertIn(".pid", cmd.drgn_option.stdout)
        self.assertIdentical(cmd.drgn_option.globals["object"], self.prog["init_task"])
        self.assertEqual(cmd.drgn_option.globals["pid"], 0)

    def test_multiple_members(self):
        cmd = self.check_crash_command("struct task_struct.pid,tgid init_task")
        self.assertRegex(cmd.stdout, r"pid = \([^)]*\)0")
        self.assertRegex(cmd.stdout, r"tgid = \([^)]*\)0")
        self.assertIn(".pid", cmd.drgn_option.stdout)
        self.assertIn(".tgid", cmd.drgn_option.stdout)
        self.assertIdentical(cmd.drgn_option.globals["object"], self.prog["init_task"])
        self.assertEqual(cmd.drgn_option.globals["pid"], 0)
        self.assertEqual(cmd.drgn_option.globals["tgid"], 0)

    @skip_unless_have_test_kmod
    def test_address_and_offset(self):
        address = hex(self.prog["drgn_test_singular_list"].next)
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry {address} -l {offset}"
        )
        self.assertIn("(struct drgn_test_list_entry){", cmd.stdout)
        self.assertIn(".value = (int)0,", cmd.stdout)
        self.assertIn(f"- {offset}", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"],
            self.prog["drgn_test_singular_list_entry"],
        )

    @skip_unless_have_test_kmod
    def test_container_of_address(self):
        address = hex(self.prog["drgn_test_singular_list"].next)
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry {address} -l drgn_test_list_entry.node"
        )
        self.assertIn("(struct drgn_test_list_entry){", cmd.stdout)
        self.assertIn(".value = (int)0,", cmd.stdout)
        self.assertIn("container_of(", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"],
            self.prog["drgn_test_singular_list_entry"],
        )

    @skip_unless_have_test_kmod
    def test_container_of_address_wrong_type(self):
        address = hex(self.prog["drgn_test_singular_list"].next)
        cmd = self.check_crash_command(
            f"struct hlist_head {address} -l drgn_test_list_entry.node"
        )
        self.assertIn("(struct hlist_head){", cmd.stdout)
        self.assertIn("container_of(", cmd.drgn_option.stdout)
        self.assertIn("reinterpret(", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"],
            reinterpret(
                "struct hlist_head", self.prog["drgn_test_singular_list_entry"]
            ),
        )

    @skip_unless_have_test_kmod
    def test_container_of_address_with_member(self):
        address = hex(self.prog["drgn_test_singular_list"].next)
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry.value {address} -l drgn_test_list_entry.node"
        )
        self.assertIn("(int)0", cmd.stdout)
        self.assertIn("container_of(", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"],
            self.prog["drgn_test_singular_list_entry"],
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_singular_list_entry"].value,
        )

    @skip_unless_have_test_kmod
    def test_positional_count(self):
        cmd = self.check_crash_command(
            "struct drgn_test_list_entry drgn_test_list_entries 3"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_list_entry){"), 3)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )

    @skip_unless_have_test_kmod
    def test_count_option(self):
        cmd = self.check_crash_command(
            "struct drgn_test_list_entry drgn_test_list_entries -c 3"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_list_entry){"), 3)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )

    @skip_unless_have_test_kmod
    def test_negative_count(self):
        address = self.prog["drgn_test_list_entries"][2].address_
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry {hex(address)} -c -3"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_list_entry){"), 3)
        self.assertIn("for object in pointer[-2:1]", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )

    @skip_unless_have_test_kmod
    def test_count_with_member(self):
        cmd = self.check_crash_command(
            "struct drgn_test_list_entry.value drgn_test_list_entries 3"
        )
        self.assertEqual(cmd.stdout.count("(int)"), 3)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIn("    value = object.value", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_list_entries"][2].value,
        )

    @skip_unless_have_test_kmod
    def test_count_with_offset(self):
        address = self.prog["drgn_test_list_entries"][0].value.address_
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "value")
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry {hex(address)} -l {offset} -c 3"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_list_entry){"), 3)
        self.assertIn(f"- {offset}", cmd.drgn_option.stdout)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )

    @skip_unless_have_test_kmod
    def test_count_with_container_of(self):
        address = self.prog["drgn_test_list_entries"][0].value.address_
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry {hex(address)} -l drgn_test_list_entry.value -c 3"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_list_entry){"), 3)
        self.assertIn("container_of(", cmd.drgn_option.stdout)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )

    @skip_unless_have_test_kmod
    def test_count_with_offset_and_member(self):
        address = self.prog["drgn_test_list_entries"][0].value.address_
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "value")
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry.value {hex(address)} -l {offset} -c 3"
        )
        self.assertEqual(cmd.stdout.count("(int)"), 3)
        self.assertIn(f"- {offset}", cmd.drgn_option.stdout)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIn("    value = object.value", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_list_entries"][2].value,
        )

    @skip_unless_have_test_kmod
    def test_count_with_container_of_and_member(self):
        address = self.prog["drgn_test_list_entries"][0].value.address_
        cmd = self.check_crash_command(
            f"struct drgn_test_list_entry.value {hex(address)} -l drgn_test_list_entry.value -c 3"
        )
        self.assertEqual(cmd.stdout.count("(int)"), 3)
        self.assertIn("container_of(", cmd.drgn_option.stdout)
        self.assertIn("for object in pointer[:3]", cmd.drgn_option.stdout)
        self.assertIn("    value = object.value", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["drgn_test_list_entries"][2]
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_list_entries"][2].value,
        )

    @skip_unless_have_test_kmod
    def test_cpuspec(self):
        cmd = self.check_crash_command(
            "struct drgn_test_percpu_struct drgn_test_percpu_structs:a"
        )
        cpus = sorted(possible_cpus())
        matches = re.findall(
            r"^\[([0-9]+)\]: [0-9a-f]+\n\(struct drgn_test_percpu_struct\)\{",
            cmd.stdout,
            flags=re.MULTILINE,
        )
        self.assertEqual([int(match) for match in matches], cpus)
        self.assertIn("per_cpu(", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["object"].cpu, max(cpus))

    @skip_unless_have_test_kmod
    def test_cpuspec_with_member(self):
        cmd = self.check_crash_command(
            "struct drgn_test_percpu_struct.cpu drgn_test_percpu_structs:a"
        )
        cpus = sorted(possible_cpus())
        matches = re.findall(r"cpu = \([^)]*\)([0-9]+)", cmd.stdout, flags=re.MULTILINE)
        self.assertEqual([int(match) for match in matches], cpus)
        self.assertIn("per_cpu(", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["object"].cpu, max(cpus))
        self.assertEqual(cmd.drgn_option.globals["cpu"], max(cpus))

    @skip_unless_have_test_kmod
    def test_cpuspec_with_count(self):
        cmd = self.check_crash_command(
            "struct drgn_test_percpu_struct drgn_test_percpu_arrays:a 3"
        )
        cpus = sorted(possible_cpus())
        matches = re.findall(
            r"(cpu|i) = \([^)]*\)([0-9]+)", cmd.stdout, flags=re.MULTILINE
        )
        expected = []
        for cpu in cpus:
            for i in range(3):
                expected.append(("cpu", str(cpu)))
                expected.append(("i", str(i)))
        self.assertEqual(matches, expected)
        self.assertEqual(cmd.drgn_option.globals["object"].cpu, max(cpus))
        self.assertEqual(cmd.drgn_option.globals["object"].i, 2)

    @skip_unless_have_test_kmod
    def test_cpuspec_with_count_and_members(self):
        cmd = self.check_crash_command(
            "struct drgn_test_percpu_struct.cpu,i drgn_test_percpu_arrays:a 3"
        )
        cpus = sorted(possible_cpus())
        matches = re.findall(
            r"^(cpu|i) = \([^)]*\)([0-9]+)", cmd.stdout, flags=re.MULTILINE
        )
        expected = []
        for cpu in cpus:
            for i in range(3):
                expected.append(("cpu", str(cpu)))
                expected.append(("i", str(i)))
        self.assertEqual(matches, expected)
        self.assertEqual(cmd.drgn_option.globals["object"].cpu, max(cpus))
        self.assertEqual(cmd.drgn_option.globals["object"].i, 2)
        self.assertEqual(cmd.drgn_option.globals["cpu"], max(cpus))
        self.assertEqual(cmd.drgn_option.globals["i"], 2)

    @skip_unless_have_test_kmod
    def test_radix(self):
        self.addCleanup(self.prog.config.pop, "crash_radix", None)

        self.run_crash_command("set radix 16")
        cmd = self.run_crash_command(
            "struct drgn_test_list_entry drgn_test_singular_list_entry"
        )
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0x0")

        cmd = self.run_crash_command(
            "struct drgn_test_list_entry drgn_test_singular_list_entry -d"
        )
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0\b")

        self.run_crash_command("set radix 10")
        cmd = self.run_crash_command(
            "struct drgn_test_list_entry drgn_test_singular_list_entry"
        )
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0\b")

        cmd = self.run_crash_command(
            "struct drgn_test_list_entry drgn_test_singular_list_entry -x"
        )
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0x0")

    def test_type_with_count(self):
        self.assertRaisesRegex(
            CommandError,
            "requires address",
            self.run_crash_command,
            "struct list_head -c 1",
        )

    def test_type_with_offset(self):
        self.assertRaisesRegex(
            CommandError,
            "requires address",
            self.run_crash_command,
            "struct list_head -l 0",
        )

    def test_invalid_structure_name(self):
        self.assertRaisesRegex(
            ValueError,
            "invalid type name",
            self.check_crash_command,
            "struct 1234",
        )

    def test_invalid_member(self):
        self.assertRaisesRegex(
            ValueError,
            "invalid member name",
            self.check_crash_command,
            "struct rq.[1]",
        )
        self.assertRaisesRegex(
            ValueError,
            "invalid member name",
            self.check_crash_command,
            "struct rq.2",
        )


class TestUnion(CrashCommandTestCase):
    @skip_unless_have_test_kmod
    def test_type(self):
        cmd = self.check_crash_command("union drgn_test_union")
        self.assertIn("union drgn_test_union {", cmd.stdout)
        self.assertRegex(cmd.stdout, "(?m)^SIZE: [0-9]+$")
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "union drgn_test_union"
        )
        self.assertIsInstance(cmd.drgn_option.globals["size"], int)

    @skip_unless_have_test_kmod
    def test_typedef(self):
        cmd = self.check_crash_command("union drgn_test_anonymous_union")
        self.assertIn("} drgn_test_anonymous_union", cmd.stdout)
        self.assertIn('prog.type("drgn_test_anonymous_union")', cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "drgn_test_anonymous_union"
        )

    def test_struct(self):
        self.assertRaises(LookupError, self.run_crash_command, "union task_struct")

    def test_anonymous_struct(self):
        self.assertRaises(LookupError, self.run_crash_command, "union atomic_t")


class TestAsterisk(CrashCommandTestCase):
    def test_struct(self):
        cmd = self.check_crash_command("*list_head")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "struct list_head"
        )

    def test_struct_with_space(self):
        cmd = self.check_crash_command("* list_head")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "struct list_head"
        )

    @skip_unless_have_test_kmod
    def test_union(self):
        cmd = self.check_crash_command("*drgn_test_union")
        self.assertIn("union drgn_test_union {", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "union drgn_test_union"
        )

    def test_struct_typedef(self):
        cmd = self.check_crash_command("*atomic_t")
        self.assertIn("} atomic_t", cmd.stdout)
        self.assertIn('prog.type("atomic_t")', cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["type"].type_name(), "atomic_t")

    @skip_unless_have_test_kmod
    def test_union_typedef(self):
        cmd = self.check_crash_command("*drgn_test_anonymous_union")
        self.assertIn("} drgn_test_anonymous_union", cmd.stdout)
        self.assertIn('prog.type("drgn_test_anonymous_union")', cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "drgn_test_anonymous_union"
        )

    def test_not_found(self):
        self.assertRaises(
            LookupError, self.run_crash_command, "*drgn_test_non_existent"
        )


class TestImplicit(CrashCommandTestCase):
    def test_struct(self):
        cmd = self.check_crash_command("list_head")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "struct list_head"
        )

    @skip_unless_have_test_kmod
    def test_union(self):
        cmd = self.check_crash_command("drgn_test_union")
        self.assertIn("union drgn_test_union {", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "union drgn_test_union"
        )

    def test_struct_typedef(self):
        cmd = self.check_crash_command("atomic_t")
        self.assertIn("} atomic_t", cmd.stdout)
        self.assertIn('prog.type("atomic_t")', cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["type"].type_name(), "atomic_t")

    @skip_unless_have_test_kmod
    def test_union_typedef(self):
        cmd = self.check_crash_command("drgn_test_anonymous_union")
        self.assertIn("} drgn_test_anonymous_union", cmd.stdout)
        self.assertIn('prog.type("drgn_test_anonymous_union")', cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["type"].type_name(), "drgn_test_anonymous_union"
        )

    def test_member(self):
        cmd = self.check_crash_command("task_struct.pid init_task")
        self.assertRegex(cmd.stdout, r"pid = \([^)]*\)0")
        self.assertIn(".pid", cmd.drgn_option.stdout)
        self.assertIdentical(cmd.drgn_option.globals["object"], self.prog["init_task"])
        self.assertEqual(cmd.drgn_option.globals["pid"], 0)

    def test_not_found(self):
        self.assertRaises(
            CommandNotFoundError, self.run_crash_command, "drgn_test_non_existent"
        )


class TestTask(CrashCommandTestCase):
    def test_no_args(self):
        self.run_crash_command("set -p")

        cmd = self.check_crash_command("task")
        self.assertIn(f"PID: {os.getpid()}", cmd.stdout)
        self.assertIn("(struct task_struct){", cmd.stdout)
        self.assertIn("(struct thread_info){", cmd.stdout)

        task = find_task(self.prog, os.getpid())
        self.assertEqual(cmd.drgn_option.globals["task"], task)
        self.assertIn("pid", cmd.drgn_option.globals)
        self.assertIn("cpu", cmd.drgn_option.globals)
        self.assertIn("command", cmd.drgn_option.globals)
        self.assertEqual(cmd.drgn_option.globals["thread_info"], task_thread_info(task))

    def test_pids(self):
        cmd = self.check_crash_command("task 1 2")
        foreach_cmd = self.check_crash_command("foreach 1 2 task", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertIn("PID: 1", c.stdout)
            self.assertIn("PID: 2", c.stdout)
            self.assertIn("(struct task_struct){", c.stdout)
            self.assertIn("(struct thread_info){", c.stdout)

        task = find_task(self.prog, 2)
        self.assertEqual(cmd.drgn_option.globals["task"], task)
        self.assertIn("pid", cmd.drgn_option.globals)
        self.assertIn("cpu", cmd.drgn_option.globals)
        self.assertIn("command", cmd.drgn_option.globals)
        self.assertEqual(cmd.drgn_option.globals["thread_info"], task_thread_info(task))

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_members(self):
        cmd = self.check_crash_command("task 1 -R on_rq,prio")
        foreach_cmd = self.check_crash_command(
            "foreach 1 task -R on_rq,prio", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            self.assertNotIn("(struct task_struct){", c.stdout)
            self.assertNotIn("(struct thread_info){", c.stdout)
            self.assertIn("on_rq =", c.stdout)
            self.assertIn("prio =", c.stdout)

        task = find_task(self.prog, 1)
        self.assertEqual(cmd.drgn_option.globals["task"], task)
        self.assertIn("pid", cmd.drgn_option.globals)
        self.assertIn("cpu", cmd.drgn_option.globals)
        self.assertIn("command", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["on_rq"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["prio"], Object)
        self.assertNotIn("thread_info", cmd.drgn_option.globals)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_members_without_R(self):
        cmd = self.check_crash_command("task 1 prio")
        self.assertNotIn("(struct task_struct){", cmd.stdout)
        self.assertNotIn("(struct thread_info){", cmd.stdout)
        self.assertIn("prio =", cmd.stdout)

        task = find_task(self.prog, 1)
        self.assertEqual(cmd.drgn_option.globals["task"], task)
        self.assertIn("pid", cmd.drgn_option.globals)
        self.assertIn("cpu", cmd.drgn_option.globals)
        self.assertIn("command", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["prio"], Object)
        self.assertNotIn("thread_info", cmd.drgn_option.globals)

    def test_members_with_and_without_R(self):
        cmd = self.check_crash_command("task -R on_rq 1 prio")
        self.assertNotIn("(struct task_struct){", cmd.stdout)
        self.assertNotIn("(struct thread_info){", cmd.stdout)
        self.assertIn("on_rq =", cmd.stdout)
        self.assertIn("prio =", cmd.stdout)

        task = find_task(self.prog, 1)
        self.assertEqual(cmd.drgn_option.globals["task"], task)
        self.assertIn("pid", cmd.drgn_option.globals)
        self.assertIn("cpu", cmd.drgn_option.globals)
        self.assertIn("command", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["on_rq"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["prio"], Object)
        self.assertNotIn("thread_info", cmd.drgn_option.globals)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import asyncio
from contextlib import contextmanager
import os
from subprocess import CalledProcessError as CalledProcessError
from typing import Any, Iterator, Tuple


async def check_call(*args: Any, **kwds: Any) -> None:
    proc = await asyncio.create_subprocess_exec(*args, **kwds)
    returncode = await proc.wait()
    if returncode != 0:
        raise CalledProcessError(returncode, args)


async def check_output(*args: Any, **kwds: Any) -> bytes:
    kwds["stdout"] = asyncio.subprocess.PIPE
    proc = await asyncio.create_subprocess_exec(*args, **kwds)
    stdout = (await proc.communicate())[0]
    if proc.returncode:
        raise CalledProcessError(proc.returncode, args)
    return stdout


async def check_output_shell(cmd: str, **kwds: Any) -> bytes:
    kwds["stdout"] = asyncio.subprocess.PIPE
    proc = await asyncio.create_subprocess_shell(cmd, **kwds)
    stdout = (await proc.communicate())[0]
    if proc.returncode:
        raise CalledProcessError(proc.returncode, cmd)
    return stdout


@contextmanager
def pipe_context() -> Iterator[Tuple[int, int]]:
    pipe_r = pipe_w = None
    try:
        pipe_r, pipe_w = os.pipe()
        yield pipe_r, pipe_w
    finally:
        if pipe_r is not None:
            os.close(pipe_r)
        if pipe_w is not None:
            os.close(pipe_w)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import collections

from drgn import NULL
from drgn.helpers import ValidationError
from drgn.helpers.linux.list import (
    hlist_empty,
    hlist_for_each,
    hlist_for_each_entry,
    list_count_nodes,
    list_empty,
    list_first_entry,
    list_first_entry_or_null,
    list_for_each,
    list_for_each_entry,
    list_for_each_entry_reverse,
    list_for_each_reverse,
    list_is_singular,
    list_last_entry,
    list_next_entry,
    list_prev_entry,
    validate_list,
    validate_list_count_nodes,
    validate_list_for_each,
    validate_list_for_each_entry,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestList(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.empty = cls.prog["drgn_test_empty_list"].address_of_()
        cls.full = cls.prog["drgn_test_full_list"].address_of_()
        cls.entries = cls.prog["drgn_test_list_entries"]
        cls.num_entries = 3
        cls.singular = cls.prog["drgn_test_singular_list"].address_of_()
        cls.singular_entry = cls.prog["drgn_test_singular_list_entry"].address_of_()
        cls.singular_node = cls.singular_entry.node.address_of_()
        cls.corrupted = cls.prog["drgn_test_corrupted_list"].address_of_()

    def node(self, n):
        return self.entries[n].node.address_of_()

    def entry(self, n):
        return self.entries[n].address_of_()

    def test_list_empty(self):
        self.assertTrue(list_empty(self.empty))
        self.assertFalse(list_empty(self.full))
        self.assertFalse(list_empty(self.singular))

    def test_list_is_singular(self):
        self.assertFalse(list_is_singular(self.empty))
        self.assertFalse(list_is_singular(self.full))
        self.assertTrue(list_is_singular(self.singular))

    def test_list_count_nodes(self):
        self.assertEqual(list_count_nodes(self.empty), 0)
        self.assertEqual(list_count_nodes(self.full), self.num_entries)
        self.assertEqual(list_count_nodes(self.singular), 1)

    def test_list_first_entry(self):
        self.assertEqual(
            list_first_entry(self.full, "struct drgn_test_list_entry", "node"),
            self.entry(0),
        )
        self.assertEqual(
            list_first_entry(self.singular, "struct drgn_test_list_entry", "node"),
            self.singular_entry,
        )

    def test_list_first_entry_or_null(self):
        self.assertEqual(
            list_first_entry_or_null(self.empty, "struct drgn_test_list_entry", "node"),
            NULL(self.prog, "struct drgn_test_list_entry *"),
        )
        self.assertEqual(
            list_first_entry_or_null(self.full, "struct drgn_test_list_entry", "node"),
            self.entry(0),
        )
        self.assertEqual(
            list_first_entry_or_null(
                self.singular, "struct drgn_test_list_entry", "node"
            ),
            self.singular_entry,
        )

    def test_list_last_entry(self):
        self.assertEqual(
            list_last_entry(self.full, "struct drgn_test_list_entry", "node"),
            self.entry(self.num_entries - 1),
        )
        self.assertEqual(
            list_last_entry(self.singular, "struct drgn_test_list_entry", "node"),
            self.singular_entry,
        )

    def test_list_next_entry(self):
        for i in range(self.num_entries - 1):
            self.assertEqual(list_next_entry(self.entry(i), "node"), self.entry(i + 1))

    def test_list_prev_entry(self):
        for i in range(1, self.num_entries):
            self.assertEqual(list_prev_entry(self.entry(i), "node"), self.entry(i - 1))

    def test_list_for_each(self):
        self.assertEqual(list(list_for_each(self.empty)), [])
        self.assertEqual(
            list(list_for_each(self.full)),
            [self.node(i) for i in range(self.num_entries)],
        )
        self.assertEqual(list(list_for_each(self.singular)), [self.singular_node])

    def test_list_for_each_reverse(self):
        self.assertEqual(list(list_for_each_reverse(self.empty)), [])
        self.assertEqual(
            list(list_for_each_reverse(self.full)),
            [self.node(i) for i in range(self.num_entries - 1, -1, -1)],
        )
        self.assertEqual(
            list(list_for_each_reverse(self.singular)), [self.singular_node]
        )

    def test_list_for_each_entry(self):
        self.assertEqual(
            list(
                list_for_each_entry("struct drgn_test_list_entry", self.empty, "node")
            ),
            [],
        )
        self.assertEqual(
            list(list_for_each_entry("struct drgn_test_list_entry", self.full, "node")),
            [self.entry(i) for i in range(self.num_entries)],
        )
        self.assertEqual(
            list(
                list_for_each_entry(
                    "struct drgn_test_list_entry", self.singular, "node"
                )
            ),
            [self.singular_entry],
        )

    def test_list_for_each_entry_reverse(self):
        self.assertEqual(
            list(
                list_for_each_entry_reverse(
                    "struct drgn_test_list_entry", self.empty, "node"
                )
            ),
            [],
        )
        self.assertEqual(
            list(
                list_for_each_entry_reverse(
                    "struct drgn_test_list_entry", self.full, "node"
                )
            ),
            [self.entry(i) for i in range(self.num_entries - 1, -1, -1)],
        )
        self.assertEqual(
            list(
                list_for_each_entry_reverse(
                    "struct drgn_test_list_entry", self.singular, "node"
                )
            ),
            [self.singular_entry],
        )

    def test_validate_list(self):
        for head in (self.empty, self.full, self.singular):
            validate_list(head)
        for head in (
            self.corrupted,
            self.prog["drgn_test_list_cycle1"].address_of_(),
            self.prog["drgn_test_list_cycle2"].address_of_(),
            self.prog["drgn_test_list_cycle3"].address_of_(),
            self.prog["drgn_test_list_self_cycle"].address_of_(),
        ):
            self.assertRaises(ValidationError, validate_list, head)

    def test_validate_list_count_nodes(self):
        self.assertEqual(validate_list_count_nodes(self.empty), 0)
        self.assertEqual(validate_list_count_nodes(self.full), self.num_entries)
        self.assertEqual(validate_list_count_nodes(self.singular), 1)
        self.assertRaises(ValidationError, validate_list_count_nodes, self.corrupted)

    def test_validate_list_for_each(self):
        for head in (self.empty, self.full, self.singular):
            self.assertEqual(
                list(validate_list_for_each(head)), list(list_for_each(head))
            )
        self.assertRaises(
            ValidationError,
            collections.deque,
            validate_list_for_each(self.corrupted),
            0,
        )

    def test_validate_list_for_each_entry(self):
        for head in (self.empty, self.full, self.singular):
            self.assertEqual(
                list(
                    validate_list_for_each_entry(
                        "struct drgn_test_list_entry", head, "node"
                    )
                ),
                list(list_for_each_entry("struct drgn_test_list_entry", head, "node")),
            )
        self.assertRaises(
            ValidationError,
            collections.deque,
            validate_list_for_each_entry(
                "struct drgn_test_list_entry", self.corrupted, "node"
            ),
            0,
        )


@skip_unless_have_test_kmod
class TestHlist(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.empty = cls.prog["drgn_test_empty_hlist"].address_of_()
        cls.full = cls.prog["drgn_test_full_hlist"].address_of_()
        cls.entries = cls.prog["drgn_test_hlist_entries"]
        cls.num_entries = 3

    def node(self, n):
        return self.entries[n].node.address_of_()

    def entry(self, n):
        return self.entries[n].address_of_()

    def test_hlist_empty(self):
        self.assertTrue(hlist_empty(self.empty))
        self.assertFalse(hlist_empty(self.full))

    def test_hlist_for_each(self):
        self.assertEqual(list(hlist_for_each(self.empty)), [])
        self.assertEqual(
            list(hlist_for_each(self.full)),
            [self.node(i) for i in range(self.num_entries)],
        )

    def test_hlist_for_each_entry(self):
        self.assertEqual(
            list(
                hlist_for_each_entry("struct drgn_test_hlist_entry", self.empty, "node")
            ),
            [],
        )
        self.assertEqual(
            list(
                hlist_for_each_entry("struct drgn_test_hlist_entry", self.full, "node")
            ),
            [self.entry(i) for i in range(self.num_entries)],
        )



# Copyright (c) 2023, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import NULL, Object
from drgn.helpers.linux.idr import idr_find, idr_for_each, idr_for_each_entry
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestIDR(LinuxKernelTestCase):
    def test_idr_find_empty(self):
        root = self.prog["drgn_test_idr_empty"].address_of_()
        self.assertIdentical(idr_find(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 100), NULL(self.prog, "void *"))

    def test_idr_for_each_empty(self):
        root = self.prog["drgn_test_idr_empty"].address_of_()
        self.assertIdentical(list(idr_for_each(root)), [])

    def test_idr_find_one(self):
        root = self.prog["drgn_test_idr_one"].address_of_()
        self.assertIdentical(idr_find(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 65), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 66), Object(self.prog, "void *", 0xDEADB00))
        self.assertIdentical(idr_find(root, 67), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 100), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 256 + 66), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 2**24 + 66), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 2**56 + 66), NULL(self.prog, "void *"))

    def test_idr_for_each_one(self):
        root = self.prog["drgn_test_idr_one"].address_of_()
        self.assertIdentical(
            list(idr_for_each(root)),
            [(66, Object(self.prog, "void *", 0xDEADB00))],
        )

    def test_idr_lookup_one_at_zero(self):
        root = self.prog["drgn_test_idr_one_at_zero"].address_of_()
        self.assertIdentical(idr_find(root, 0), Object(self.prog, "void *", 0x1234))
        self.assertIdentical(idr_find(root, 1), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 100), NULL(self.prog, "void *"))

    def test_idr_for_each_one_at_zero(self):
        root = self.prog["drgn_test_idr_one_at_zero"].address_of_()
        self.assertIdentical(
            list(idr_for_each(root)), [(0, Object(self.prog, "void *", 0x1234))]
        )

    def test_idr_find_sparse(self):
        root = self.prog["drgn_test_idr_sparse"].address_of_()
        self.assertIdentical(idr_find(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 1), Object(self.prog, "void *", 0x1234))
        self.assertIdentical(idr_find(root, 2), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 0x40), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 0x70), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 0x7F), NULL(self.prog, "void *"))
        self.assertIdentical(idr_find(root, 0x80), Object(self.prog, "void *", 0x5678))
        self.assertIdentical(idr_find(root, 0xEE), Object(self.prog, "void *", 0x9ABC))
        self.assertIdentical(idr_find(root, 0xEF), NULL(self.prog, "void *"))

    def test_idr_for_each_sparse(self):
        root = self.prog["drgn_test_idr_sparse"].address_of_()
        self.assertIdentical(
            list(idr_for_each(root)),
            [
                (1, Object(self.prog, "void *", 0x1234)),
                (0x80, Object(self.prog, "void *", 0x5678)),
                (0xEE, Object(self.prog, "void *", 0x9ABC)),
            ],
        )

    def test_idr_for_each_entry(self):
        root = self.prog["drgn_test_idr_sparse"].address_of_()
        self.assertIdentical(
            list(idr_for_each_entry(root, "int")),
            [
                (1, Object(self.prog, "int *", 0x1234)),
                (0x80, Object(self.prog, "int *", 0x5678)),
                (0xEE, Object(self.prog, "int *", 0x9ABC)),
            ],
        )



# Copyright (c) 2024 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
import re
import tempfile
from unittest import TestCase

from drgn import Symbol, SymbolBinding, SymbolKind
from drgn.helpers.linux.kallsyms import (
    _load_builtin_kallsyms,
    _load_proc_kallsyms,
    load_module_kallsyms,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


def compare_local_symbols(self, finder, modules=False):
    expr = re.compile(
        r"(?P<address>[0-9a-f]+) (?P<kind>.) " r"(?P<name>[^\s]+)(\s+(?P<mod>\[\w+\]))?"
    )
    names = {}
    count = 0
    with open("/proc/kallsyms") as f:
        for line in f:
            match = expr.fullmatch(line.strip())
            self.assertIsNotNone(match, line)
            if match.group("mod") and not modules:
                break
            count += 1
            name = match.group("name")
            addr = int(match.group("address"), 16)
            names.setdefault(name, []).append((addr, match.group("kind"), name))

    for name, syms in names.items():
        res = finder(None, name, None, False)
        expected_addrs = sorted(t[0] for t in syms)
        found_addrs = sorted(s.address for s in res)
        self.assertEqual(expected_addrs, found_addrs)

    all_res = finder(None, None, None, False)
    self.assertEqual(count, len(all_res))


KALLSYMS_DATA = b"""\
0000000000000000 u null
0000000000000008 d local_data
0000000000000010 B global_bss
0000000000000020 v weak_symbol
0000000000000040 ? unknown
0000000000001000 T text [mymod]
0000000000002000 T modfunc1 [mymod2]
0000000000002010 T modfunc2 [mymod2]
0000000000002008 T modfunc3 [mymod2]
"""


class TestProcKallsyms(TestCase):
    def test_static_data(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(KALLSYMS_DATA)
            f.flush()
            finder = _load_proc_kallsyms(filename=f.name)

        syms = finder(None, None, None, False)
        expected = [
            Symbol("null", 0x0, 8, SymbolBinding.UNIQUE, SymbolKind.UNKNOWN),
            Symbol("local_data", 0x8, 8, SymbolBinding.UNKNOWN, SymbolKind.OBJECT),
            Symbol("global_bss", 0x10, 16, SymbolBinding.GLOBAL, SymbolKind.OBJECT),
            Symbol("weak_symbol", 0x20, 32, SymbolBinding.WEAK, SymbolKind.OBJECT),
            # this one has zero size since it is at the end of vmlinux
            Symbol("unknown", 0x40, 0, SymbolBinding.UNKNOWN, SymbolKind.UNKNOWN),
        ]
        self.assertEqual(syms, expected)

    def test_static_data_with_modules(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(KALLSYMS_DATA)
            f.flush()
            finder = _load_proc_kallsyms(filename=f.name, modules=True)

        syms = finder(None, None, None, False)
        expected = [
            Symbol("null", 0x0, 8, SymbolBinding.UNIQUE, SymbolKind.UNKNOWN),
            Symbol("local_data", 0x8, 8, SymbolBinding.UNKNOWN, SymbolKind.OBJECT),
            Symbol("global_bss", 0x10, 16, SymbolBinding.GLOBAL, SymbolKind.OBJECT),
            Symbol("weak_symbol", 0x20, 32, SymbolBinding.WEAK, SymbolKind.OBJECT),
            # this one has zero size since it is at the end of vmlinux
            Symbol("unknown", 0x40, 0, SymbolBinding.UNKNOWN, SymbolKind.UNKNOWN),
            # this one has zero size since it is at the end of a module
            Symbol("text", 0x1000, 0, SymbolBinding.GLOBAL, SymbolKind.FUNC),
            # this one has a non-zero size since it is within a module
            Symbol("modfunc1", 0x2000, 16, SymbolBinding.GLOBAL, SymbolKind.FUNC),
            # this one has a zero size since it is at the end of the file. It is
            # returned in sorted order by address despite kallsyms not
            # containing it in sorted order.
            Symbol("modfunc3", 0x2008, 0, SymbolBinding.GLOBAL, SymbolKind.FUNC),
            # this one has a zero size since it is followed by an out-of-order
            # symbol
            Symbol("modfunc2", 0x2010, 0, SymbolBinding.GLOBAL, SymbolKind.FUNC),
        ]
        self.assertEqual(syms, expected)


class TestBuiltinKallsyms(LinuxKernelTestCase):
    def test_builtin_kallsyms(self):
        if b"kallsyms_num_syms" not in self.prog["VMCOREINFO"].string_():
            self.skipTest("VMCOREINFO is missing necessary symbols")
        finder = _load_builtin_kallsyms(self.prog)
        compare_local_symbols(self, finder)

    @skip_unless_have_test_kmod
    def test_module_kallsyms(self):
        finder = load_module_kallsyms(self.prog)
        test_data = finder(None, "drgn_test_empty_list", None, True)[0]
        self.assertEqual("drgn_test_empty_list", test_data.name)
        self.assertEqual(SymbolKind.OBJECT, test_data.kind)
        self.assertIn(test_data.binding, (SymbolBinding.GLOBAL, SymbolBinding.UNKNOWN))
        size = self.prog.type("struct list_head").size
        self.assertEqual(size, test_data.size)
        address = self.prog.object("drgn_test_empty_list").address_
        self.assertEqual(address, test_data.address)



# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.swait import swait_active, swait_for_each_task
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestSimpleWaitqueue(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.swaitq = cls.prog["drgn_test_swaitq"].address_of_()
        cls.empty_swaitq = cls.prog["drgn_test_empty_swaitq"].address_of_()

    def test_swait_active(self):
        self.assertTrue(swait_active(self.swaitq))
        self.assertFalse(swait_active(self.empty_swaitq))

    def test_swait_for_each_task(self):
        self.assertEqual(list(swait_for_each_task(self.empty_swaitq)), [])
        self.assertEqual(
            list(swait_for_each_task(self.swaitq)),
            [self.prog["drgn_test_swaitq_kthread"]],
        )



# Copyright 2017-2019 by Takeshi KOMIYA
# SPDX-License-Identifier: Apache-2.0
# From https://pypi.org/project/sphinxcontrib-details-directive/, patched to
# use the proper name for the :class: option.

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.nodes import NodeMatcher


class details(nodes.Element, nodes.General):
    pass


class summary(nodes.TextElement, nodes.General):
    pass


def visit_details(self, node):
    if node.get('opened'):
        self.body.append(self.starttag(node, 'details', open="open"))
    else:
        self.body.append(self.starttag(node, 'details'))


def depart_details(self, node):
    self.body.append('</details>')


def visit_summary(self, node):
    self.body.append(self.starttag(node, 'summary'))


def depart_summary(self, node):
    self.body.append('</summary>')


class DetailsDirective(Directive):
    required_arguments = 1
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'class': directives.class_option,
        'name': directives.unchanged,
        'open': directives.flag,
    }

    def run(self):
        admonition = nodes.container('',
                                     classes=self.options.get('class', []),
                                     opened='open' in self.options,
                                     type='details')
        textnodes, messages = self.state.inline_text(self.arguments[0],
                                                     self.lineno)
        admonition += nodes.paragraph(self.arguments[0], '', *textnodes)
        admonition += messages
        self.state.nested_parse(self.content, self.content_offset, admonition)
        self.add_name(admonition)
        return [admonition]


class DetailsTransform(SphinxPostTransform):
    default_priority = 200
    builders = ('html',)

    def run(self):
        matcher = NodeMatcher(nodes.container, type='details')
        for node in self.document.traverse(matcher):
            newnode = details(**node.attributes)
            newnode += summary('', '', *node[0])
            newnode.extend(node[1:])
            node.replace_self(newnode)


def setup(app):
    app.add_node(details, html=(visit_details, depart_details))
    app.add_node(summary, html=(visit_summary, depart_summary))
    app.add_directive('details', DetailsDirective)
    app.add_post_transform(DetailsTransform)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ctypes

import _drgn

_drgn_cdll = ctypes.CDLL(_drgn.__file__)


class _drgn_error(ctypes.Structure):
    _fields_ = [
        ("code", ctypes.c_uint),
        ("errnum", ctypes.c_int),
        ("path", ctypes.c_char_p),
        ("msg", ctypes.c_char_p),
    ]


class _drgn_type(ctypes.Structure):
    pass


class _drgn_qualified_type(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.POINTER(_drgn_type)),
        ("qualifiers", ctypes.c_uint),
    ]


class _drgn_token(ctypes.Structure):
    _fields_ = [
        ("kind", ctypes.c_int),
        ("value", ctypes.c_void_p),
        ("len", ctypes.c_size_t),
    ]


_drgn_cdll.drgn_test_serialize_bits.restype = None
_drgn_cdll.drgn_test_serialize_bits.argtypes = [
    ctypes.c_void_p,
    ctypes.c_uint64,
    ctypes.c_uint64,
    ctypes.c_uint8,
    ctypes.c_bool,
]
_drgn_cdll.drgn_test_deserialize_bits.restype = ctypes.c_uint64
_drgn_cdll.drgn_test_deserialize_bits.argtypes = [
    ctypes.c_void_p,
    ctypes.c_uint64,
    ctypes.c_uint8,
    ctypes.c_bool,
]


def serialize_bits(buf, bit_offset, uvalue, bit_size, little_endian):
    assert (bit_offset + bit_size + 7) // 8 <= len(buf)
    c_buf = (ctypes.c_char * len(buf)).from_buffer(buf)
    return _drgn_cdll.drgn_test_serialize_bits(
        c_buf, bit_offset, uvalue, bit_size, little_endian
    )


def deserialize_bits(buf, bit_offset, bit_size, little_endian):
    assert (bit_offset + bit_size + 7) // 8 <= len(buf)
    c_buf = (ctypes.c_char * len(buf)).from_buffer_copy(buf)
    return _drgn_cdll.drgn_test_deserialize_bits(
        c_buf, bit_offset, bit_size, little_endian
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os
from pathlib import Path
import tempfile
import unittest

from drgn import NULL, cast
from drgn.helpers.linux.cgroup import (
    cgroup_get_from_path,
    cgroup_name,
    cgroup_parent,
    cgroup_path,
    css_for_each_child,
    css_for_each_descendant_pre,
    sock_cgroup_ptr,
)
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    LinuxKernelTestCase,
    create_socket,
    fork_and_stop,
    iter_mounts,
)


@contextlib.contextmanager
def tmp_cgroup():
    for mnt in iter_mounts():
        if mnt.fstype == "cgroup2":
            break
    else:
        raise unittest.SkipTest("cgroup2 not mounted")

    path = Path(tempfile.mkdtemp(prefix="drgn-tests-", dir=mnt.mount_point))
    try:
        yield path
    finally:
        path.rmdir()


@contextlib.contextmanager
def tmp_cgroups():
    with tmp_cgroup() as parent:
        child = parent / "child"
        child.mkdir()
        try:
            yield parent, child
        finally:
            child.rmdir()


class TestCgroup(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        parent_cgroup_dir, child_cgroup_dir = cls.enterClassContext(tmp_cgroups())

        cls.root_cgroup = cls.prog["cgrp_dfl_root"].cgrp.address_of_()

        with fork_and_stop() as pid:
            task = find_task(cls.prog, pid)

            cls.parent_cgroup_name = os.fsencode(parent_cgroup_dir.name)
            cls.parent_cgroup_path = b"/" + cls.parent_cgroup_name

            (parent_cgroup_dir / "cgroup.procs").write_text(str(pid))
            cls.parent_cgroup = task.cgroups.dfl_cgrp.read_()

            cls.child_cgroup_name = os.fsencode(child_cgroup_dir.name)
            cls.child_cgroup_path = (
                cls.parent_cgroup_path + b"/" + cls.child_cgroup_name
            )

            (child_cgroup_dir / "cgroup.procs").write_text(str(pid))
            cls.child_cgroup = task.cgroups.dfl_cgrp.read_()

    def test_cgroup_parent(self):
        self.assertEqual(cgroup_parent(self.child_cgroup), self.parent_cgroup)
        self.assertEqual(cgroup_parent(self.parent_cgroup), self.root_cgroup)
        self.assertEqual(
            cgroup_parent(self.root_cgroup), NULL(self.prog, "struct cgroup *")
        )

    def test_cgroup_name(self):
        self.assertEqual(cgroup_name(self.root_cgroup), b"/")
        self.assertEqual(cgroup_name(self.parent_cgroup), self.parent_cgroup_name)
        self.assertEqual(cgroup_name(self.child_cgroup), self.child_cgroup_name)

    def test_cgroup_path(self):
        self.assertEqual(cgroup_path(self.root_cgroup), b"/")
        self.assertEqual(cgroup_path(self.parent_cgroup), self.parent_cgroup_path)
        self.assertEqual(cgroup_path(self.child_cgroup), self.child_cgroup_path)

    def test_cgroup_get_from_path(self):
        self.assertEqual(cgroup_get_from_path(self.prog, "/"), self.root_cgroup)
        self.assertEqual(
            cgroup_get_from_path(self.prog, self.parent_cgroup_path), self.parent_cgroup
        )
        self.assertEqual(
            cgroup_get_from_path(self.prog, self.child_cgroup_path), self.child_cgroup
        )
        self.assertEqual(
            cgroup_get_from_path(self.prog, self.parent_cgroup_path + b"/foo"),
            NULL(self.prog, "struct cgroup *"),
        )

    def test_cgroup_socket(self):
        with create_socket() as sock:
            task = find_task(self.prog, os.getpid())
            file = fget(task, sock.fileno())
            sk = cast("struct socket *", file.private_data).sk
            self.assertEqual(sock_cgroup_ptr(sk.sk_cgrp_data), task.cgroups.dfl_cgrp)

    @staticmethod
    def _cgroup_iter_paths(fn, cgroup):
        return [cgroup_path(css.cgroup) for css in fn(cgroup.self.address_of_())]

    def test_css_for_each_child(self):
        children = self._cgroup_iter_paths(css_for_each_child, self.root_cgroup)
        self.assertIn(self.parent_cgroup_path, children)
        self.assertNotIn(self.child_cgroup_path, children)

        self.assertEqual(
            self._cgroup_iter_paths(css_for_each_child, self.parent_cgroup),
            [self.child_cgroup_path],
        )

        self.assertEqual(
            self._cgroup_iter_paths(css_for_each_child, self.child_cgroup), []
        )

    def test_css_for_each_descendant_pre(self):
        descendants = self._cgroup_iter_paths(
            css_for_each_descendant_pre, self.root_cgroup
        )
        self.assertEqual(descendants[0], b"/")
        self.assertIn(self.parent_cgroup_path, descendants)
        self.assertIn(self.child_cgroup_path, descendants)
        self.assertLess(
            descendants.index(self.parent_cgroup_path),
            descendants.index(self.child_cgroup_path),
        )

        self.assertEqual(
            self._cgroup_iter_paths(css_for_each_descendant_pre, self.parent_cgroup),
            [self.parent_cgroup_path, self.child_cgroup_path],
        )

        self.assertEqual(
            self._cgroup_iter_paths(css_for_each_descendant_pre, self.child_cgroup),
            [self.child_cgroup_path],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from collections import namedtuple


def _append_uleb128(buf, value):
    assert value >= 0
    while True:
        byte = value & 0x7F
        value >>= 7
        if value:
            buf.append(byte | 0x80)
        else:
            buf.append(byte)
            break


def _append_sleb128(buf, value):
    while True:
        byte = value & 0x7F
        value >>= 7
        if (not value and not (byte & 0x40)) or (value == -1 and (byte & 0x40)):
            buf.append(byte)
            break
        else:
            buf.append(byte | 0x80)


U8 = namedtuple("U8", ["value"])
U8._append = lambda self, buf, byteorder: buf.append(self.value)
S8 = namedtuple("S8", ["value"])
S8._append = lambda self, buf, byteorder: buf.append(self.value & 0xFF)
U16 = namedtuple("U16", ["value"])
U16._append = lambda self, buf, byteorder: buf.extend(self.value.to_bytes(2, byteorder))
S16 = namedtuple("S16", ["value"])
S16._append = lambda self, buf, byteorder: buf.extend(
    self.value.to_bytes(2, byteorder, signed=True)
)
U32 = namedtuple("U32", ["value"])
U32._append = lambda self, buf, byteorder: buf.extend(self.value.to_bytes(4, byteorder))
S32 = namedtuple("S32", ["value"])
S32._append = lambda self, buf, byteorder: buf.extend(
    self.value.to_bytes(4, byteorder, signed=True)
)
U64 = namedtuple("U64", ["value"])
U64._append = lambda self, buf, byteorder: buf.extend(self.value.to_bytes(8, byteorder))
S64 = namedtuple("S64", ["value"])
S64._append = lambda self, buf, byteorder: buf.extend(
    self.value.to_bytes(8, byteorder, signed=True)
)
ULEB128 = namedtuple("ULEB128", ["value"])
ULEB128._append = lambda self, buf, byteorder: _append_uleb128(buf, self.value)
SLEB128 = namedtuple("SLEB128", ["value"])
SLEB128._append = lambda self, buf, byteorder: _append_sleb128(buf, self.value)


def assemble(*args, little_endian=True):
    byteorder = "little" if little_endian else "big"
    buf = bytearray()
    for arg in args:
        if isinstance(arg, bytes):
            buf.extend(arg)
        else:
            arg._append(buf, byteorder)
    return buf



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ctypes
import os
import re
import subprocess
import sys

from _drgn_util.platform import NORMALIZED_MACHINE_NAME, SYS

KEXEC_FILE_ON_CRASH = 2
KEXEC_FILE_NO_INITRAMFS = 4


def main() -> None:
    with open("/proc/cmdline", "rb") as f:
        cmdline = f.read().rstrip(b"\n")
        cmdline = re.sub(rb"(^|\s)crashkernel=\S+", b"", cmdline)
        # `nokaslr` is required to avoid sporadically failing to reserve space for the
        # capture kernel
        cmdline += b" nokaslr"
        if os.getenv("KDUMP_NEEDS_NOSMP"):
            # `nosmp` is required to avoid QEMU sporadically failing an internal
            # assertion when using emulation.
            cmdline += b" nosmp"

    # On RISC-V, kexec_file_load() is supported for Image (vmlinuz) files only
    # since Linux kernel commit 809a11eea8e8 ("riscv: kexec_file: Support
    # loading Image binary file") (in v6.16), but for ELF files since Linux
    # kernel commit 6261586e0c91 ("RISC-V: Add kexec_file support") (in v5.19).
    if NORMALIZED_MACHINE_NAME == "riscv64":
        kernel_image = f"/lib/modules/{os.uname().release}/build/vmlinux"
    else:
        kernel_image = f"/lib/modules/{os.uname().release}/vmlinuz"

    # On x86-64 and RISC-V, kexec_file_load() is supported on all kernel
    # versions we care about, and it's simple enough to call ourselves. On
    # other architectures, we just use kexec(8).
    if NORMALIZED_MACHINE_NAME in {"x86_64", "riscv64"}:
        syscall = ctypes.CDLL(None, use_errno=True).syscall
        syscall.restype = ctypes.c_long

        with open(kernel_image, "rb") as kernel:
            if syscall(
                ctypes.c_long(SYS["kexec_file_load"]),
                ctypes.c_int(kernel.fileno()),
                ctypes.c_int(-1),
                ctypes.c_ulong(len(cmdline) + 1),
                ctypes.c_char_p(cmdline + b"\0"),
                ctypes.c_ulong(KEXEC_FILE_ON_CRASH | KEXEC_FILE_NO_INITRAMFS),
            ):
                errno = ctypes.get_errno()
                raise OSError(errno, os.strerror(errno))
    else:
        subprocess.check_call(
            [
                "kexec",
                "--load-panic",
                "--kexec-syscall-auto",
                "--command-line=" + cmdline.decode(),
                kernel_image,
            ]
        )

    with open("/proc/self/comm", "w") as f:
        f.write("selfdestruct")

    # Avoid panicking from CPU 0 on s390x. See _skip_if_cpu0_on_s390x().
    if NORMALIZED_MACHINE_NAME == "s390x":
        cpus = os.sched_getaffinity(0)
        cpus.remove(0)
        if cpus:
            os.sched_setaffinity(0, cpus)

    # Try the drgn_test kmod crash method first.
    try:
        with open("/sys/kernel/drgn_test/crash", "w") as f:
            f.write("1")
    except FileNotFoundError:
        pass

    # Fall back to sysrq-trigger.
    with open("/proc/sysrq-trigger", "w") as f:
        f.write("c")


if __name__ == "__main__":
    if os.path.exists("/proc/vmcore") and len(sys.argv) > 1:
        # We're in the kdump kernel already. Assume the user wanted to run
        # a command within it.
        os.execvp(sys.argv[1], sys.argv[1:])
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from _drgn_util.multilinewrap import multiline_fill, multiline_wrap
from tests import TestCase


class TestMultilineWrap(TestCase):
    def assert_wrap(self, text, lines, *, width, indent=""):
        self.assertEqual(multiline_wrap(text, width, indent=indent), lines)

    def test_one_line(self):
        self.assert_wrap(
            "Sphinx of black quartz, judge my vow.",
            [
                "Sphinx of black quartz,",
                "judge my vow.",
            ],
            width=25,
        )

    def test_one_paragraph(self):
        self.assert_wrap(
            "Sphinx of black\nquartz, judge\nmy vow.",
            [
                "Sphinx of black quartz,",
                "judge my vow.",
            ],
            width=25,
        )

    def test_dedent(self):
        self.assert_wrap(
            """
            Sphinx of black
            quartz, judge
            my vow.
            """,
            [
                "Sphinx of black quartz,",
                "judge my vow.",
            ],
            width=25,
        )

    def test_one_character_words(self):
        self.assert_wrap("A\nB\nC", ["A B C"], width=25)

    def test_leading_whitespace(self):
        self.assert_wrap("A\n B\nC", ["A  B C"], width=25)

    def test_multiple_paragraphs(self):
        self.assert_wrap(
            """
            Sphinx of black quartz, judge my vow.

            The quick brown fox jumps over the lazy dog.
            """,
            [
                "Sphinx of black quartz,",
                "judge my vow.",
                "",
                "The quick brown fox jumps",
                "over the lazy dog.",
            ],
            width=25,
        )

    def test_multiple_paragraphs_extra_whitespace(self):
        self.assert_wrap(
            """

            Sphinx of black
            quartz, judge
            my vow.


            The quick brown
            fox jumps over
            the lazy dog.

            """,
            [
                "Sphinx of black quartz,",
                "judge my vow.",
                "",
                "The quick brown fox jumps",
                "over the lazy dog.",
            ],
            width=25,
        )

    def test_line_block(self):
        self.assert_wrap(
            """
            |A B C
            |      D E F
            |G H I
            """,
            [
                "A B C",
                "      D E F",
                "G H I",
            ],
            width=1,
        )

    def test_line_block_whitespace(self):
        self.assert_wrap(
            # Use "\x20" to avoid trailing whitespace warnings.
            """
            |A B C
            |\x20
            |      D E F
            |
            |G H I
            """,
            [
                "A B C",
                " ",
                "      D E F",
                "",
                "G H I",
            ],
            width=1,
        )

    def test_line_block_and_paragraphs(self):
        self.assert_wrap(
            """
            Sphinx of black quartz, judge my vow.

            |A B C
            |      D E F
            |G H I

            The quick brown fox jumps over the lazy dog.
            """,
            [
                "Sphinx of",
                "black",
                "quartz,",
                "judge my",
                "vow.",
                "",
                "A B C",
                "      D E F",
                "G H I",
                "",
                "The quick",
                "brown fox",
                "jumps over",
                "the lazy",
                "dog.",
            ],
            width=10,
        )

    def test_line_block_in_paragraph(self):
        self.assert_wrap(
            """
            Sphinx of black quartz, judge my vow.
            |A B C
            |      D E F
            |G H I
            The quick brown fox jumps over the lazy dog.
            """,
            [
                "Sphinx of",
                "black",
                "quartz,",
                "judge my",
                "vow.",
                "A B C",
                "      D E F",
                "G H I",
                "The quick",
                "brown fox",
                "jumps over",
                "the lazy",
                "dog.",
            ],
            width=10,
        )

    def test_fill(self):
        self.assertEqual(
            multiline_fill("Sphinx of black\nquartz, judge\nmy vow.", 25),
            "Sphinx of black quartz,\njudge my vow.",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import unittest.mock

from drgn import NoDefaultProgramError, Object, get_default_prog, set_default_prog
from drgn.helpers.common.prog import (
    takes_object_or_program_or_default,
    takes_program_or_default,
)
from tests import IdenticalMatcher, TestCase, mock_program


@contextlib.contextmanager
def mock_default_prog():
    try:
        old_default_prog = get_default_prog()
    except NoDefaultProgramError:
        old_default_prog = None
    prog = mock_program()
    try:
        set_default_prog(prog)
        yield prog
    finally:
        set_default_prog(old_default_prog)


class TestTakesProgramOrDefault(TestCase):
    def setUp(self):
        def f1(prog):
            pass

        self.mock1 = unittest.mock.create_autospec(f1)

        def f2(prog, x):
            pass

        self.mock2 = unittest.mock.create_autospec(f2)

        def f3(prog, x, s):
            pass

        self.mock3 = unittest.mock.create_autospec(f3)

    def test_explicit_prog_no_args(self):
        prog = mock_program()
        takes_program_or_default(self.mock1)(prog)
        self.mock1.assert_called_once_with(prog)

    def test_explicit_prog_kwarg_no_args(self):
        prog = mock_program()
        takes_program_or_default(self.mock1)(prog=prog)
        self.mock1.assert_called_once_with(prog=prog)

    def test_default_prog_no_args(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock1)()
            self.mock1.assert_called_once_with(prog)

    def test_explicit_prog_one_arg(self):
        prog = mock_program()
        takes_program_or_default(self.mock2)(prog, 1)
        self.mock2.assert_called_once_with(prog, 1)

    def test_explicit_prog_one_kwarg(self):
        prog = mock_program()
        takes_program_or_default(self.mock2)(prog, x=1)
        self.mock2.assert_called_once_with(prog, x=1)

    def test_explicit_prog_kwarg_one_kwarg(self):
        prog = mock_program()
        takes_program_or_default(self.mock2)(prog=prog, x=1)
        self.mock2.assert_called_once_with(prog=prog, x=1)

    def test_object_prog_one_arg(self):
        prog = mock_program()
        takes_program_or_default(self.mock2)(Object(prog, "int", 1))
        self.mock2.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1))
        )

    def test_object_prog_one_kwarg(self):
        prog = mock_program()
        takes_program_or_default(self.mock2)(x=Object(prog, "int", 1))
        self.mock2.assert_called_once_with(
            prog, x=IdenticalMatcher(Object(prog, "int", 1))
        )

    def test_default_prog_one_arg(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock2)(1)
            self.mock2.assert_called_once_with(prog, 1)

    def test_default_prog_one_kwarg(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock2)(x=1)
            self.mock2.assert_called_once_with(prog, x=1)

    def test_explicit_prog_two_args(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(prog, 1, "foo")
        self.mock3.assert_called_once_with(prog, 1, "foo")

    def test_explicit_prog_one_arg_one_kwarg(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(prog, 1, s="foo")
        self.mock3.assert_called_once_with(prog, 1, s="foo")

    def test_explicit_prog_two_kwargs(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(prog, x=1, s="foo")
        self.mock3.assert_called_once_with(prog, x=1, s="foo")

    def test_explicit_prog_kwarg_two_kwargs(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(prog=prog, x=1, s="foo")
        self.mock3.assert_called_once_with(prog=prog, x=1, s="foo")

    def test_object_prog_two_args(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(Object(prog, "int", 2), "foo")
        self.mock3.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 2)), "foo"
        )

    def test_object_prog_one_arg_one_kwarg(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(Object(prog, "int", 2), s="foo")
        self.mock3.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 2)), s="foo"
        )

    def test_object_prog_two_kwargs(self):
        prog = mock_program()
        takes_program_or_default(self.mock3)(x=Object(prog, "int", 2), s="foo")
        self.mock3.assert_called_once_with(
            prog, x=IdenticalMatcher(Object(prog, "int", 2)), s="foo"
        )

    def test_default_prog_two_args(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock3)(2, "foo")
            self.mock3.assert_called_once_with(prog, 2, "foo")

    def test_default_prog_one_arg_one_kwarg(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock3)(2, s="foo")
            self.mock3.assert_called_once_with(prog, 2, s="foo")

    def test_default_prog_two_kwargs(self):
        with mock_default_prog() as prog:
            takes_program_or_default(self.mock3)(x=2, s="foo")
            self.mock3.assert_called_once_with(prog, x=2, s="foo")


class TestTakesObjectOrProgramOrDefault(TestCase):
    def setUp(self):
        def f1(prog, obj):
            pass

        self.mock1 = unittest.mock.create_autospec(f1)

        def f2(prog, obj, x):
            pass

        self.mock2 = unittest.mock.create_autospec(f2)

        def f3(prog, obj, x, s):
            pass

        self.mock3 = unittest.mock.create_autospec(f3)

    def test_explicit_prog_no_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock1)(prog)
        self.mock1.assert_called_once_with(prog, None)

    def test_explicit_prog_kwarg_no_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock1)(prog=prog)
        self.mock1.assert_called_once_with(prog=prog, obj=None)

    def test_obj_no_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock1)(Object(prog, "int", 1))
        self.mock1.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1))
        )

    def test_obj_kwarg_no_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock1)(obj=Object(prog, "int", 1))
        self.mock1.assert_called_once_with(
            prog, obj=IdenticalMatcher(Object(prog, "int", 1))
        )

    def test_default_prog_no_args(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock1)()
            self.mock1.assert_called_once_with(prog, None)

    def test_explicit_prog_one_arg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(prog, 1)
        self.mock2.assert_called_once_with(prog, None, 1)

    def test_explicit_prog_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(prog, x=1)
        self.mock2.assert_called_once_with(prog, None, x=1)

    def test_explicit_prog_kwarg_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(prog=prog, x=1)
        self.mock2.assert_called_once_with(prog=prog, obj=None, x=1)

    def test_obj_one_arg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(Object(prog, "int", 1), 2)
        self.mock2.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1)), 2
        )

    def test_obj_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(Object(prog, "int", 1), x=2)
        self.mock2.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1)), x=2
        )

    def test_obj_kwarg_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(obj=Object(prog, "int", 1), x=2)
        self.mock2.assert_called_once_with(
            prog, obj=IdenticalMatcher(Object(prog, "int", 1)), x=2
        )

    def test_object_prog_one_arg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(Object(prog, "int", 2))
        self.mock2.assert_called_once_with(
            prog, None, IdenticalMatcher(Object(prog, "int", 2))
        )

    def test_object_prog_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock2)(x=Object(prog, "int", 2))
        self.mock2.assert_called_once_with(
            prog, None, x=IdenticalMatcher(Object(prog, "int", 2))
        )

    def test_default_prog_one_arg(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock2)(2)
            self.mock2.assert_called_once_with(prog, None, 2)

    def test_default_prog_one_kwarg(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock2)(x=2)
            self.mock2.assert_called_once_with(prog, None, x=2)

    def test_explicit_prog_two_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(prog, 1, "foo")
        self.mock3.assert_called_once_with(prog, None, 1, "foo")

    def test_explicit_prog_one_arg_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(prog, 1, s="foo")
        self.mock3.assert_called_once_with(prog, None, 1, s="foo")

    def test_explicit_prog_two_kwargs(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(prog, x=1, s="foo")
        self.mock3.assert_called_once_with(prog, None, x=1, s="foo")

    def test_explicit_prog_kwarg_two_kwargs(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(prog=prog, x=1, s="foo")
        self.mock3.assert_called_once_with(prog=prog, obj=None, x=1, s="foo")

    def test_obj_two_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(Object(prog, "int", 1), 2, "foo")
        self.mock3.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1)), 2, "foo"
        )

    def test_obj_one_arg_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(
            Object(prog, "int", 1), 2, s="foo"
        )
        self.mock3.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1)), 2, s="foo"
        )

    def test_obj_two_kwargs(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(
            Object(prog, "int", 1), x=2, s="foo"
        )
        self.mock3.assert_called_once_with(
            prog, IdenticalMatcher(Object(prog, "int", 1)), x=2, s="foo"
        )

    def test_obj_kwarg_two_kwargs(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(
            obj=Object(prog, "int", 1), x=2, s="foo"
        )
        self.mock3.assert_called_once_with(
            prog, obj=IdenticalMatcher(Object(prog, "int", 1)), x=2, s="foo"
        )

    def test_object_prog_two_args(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(Object(prog, "int", 2), "foo")
        self.mock3.assert_called_once_with(
            prog, None, IdenticalMatcher(Object(prog, "int", 2)), "foo"
        )

    def test_object_prog_one_arg_one_kwarg(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(Object(prog, "int", 2), s="foo")
        self.mock3.assert_called_once_with(
            prog, None, IdenticalMatcher(Object(prog, "int", 2)), s="foo"
        )

    def test_object_prog_two_kwargs(self):
        prog = mock_program()
        takes_object_or_program_or_default(self.mock3)(
            x=Object(prog, "int", 2), s="foo"
        )
        self.mock3.assert_called_once_with(
            prog, None, x=IdenticalMatcher(Object(prog, "int", 2)), s="foo"
        )

    def test_default_prog_two_args(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock3)(2, "foo")
            self.mock3.assert_called_once_with(prog, None, 2, "foo")

    def test_default_prog_one_arg_one_kwarg(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock3)(2, s="foo")
            self.mock3.assert_called_once_with(prog, None, 2, s="foo")

    def test_default_prog_two_kwargs(self):
        with mock_default_prog() as prog:
            takes_object_or_program_or_default(self.mock3)(x=2, s="foo")
            self.mock3.assert_called_once_with(prog, None, x=2, s="foo")
